# Generated by Django 3.1 on 2020-08-14 19:52

from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='WSPayRequest',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cart_id', models.PositiveIntegerField()),
                ('status', models.CharField(choices=[('PENDING', 'pending'), ('COMPLETED', 'completed'), ('CANCELLED', 'cancelled'), ('FAILED', 'failed')], default='PENDING', max_length=15)),
                ('request_uuid', models.UUIDField(default=uuid.uuid4)),
                ('response', models.TextField(null=True)),
                ('additional_data', models.TextField(blank=True, help_text='Use this to store any data you want to preserve when making a request')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
    ]



from django.conf import settings
from django.urls import path, re_path

from wspay.views import ProcessView, ProcessResponseView, TestView, TransactionReportView

app_name = 'wspay'

urlpatterns = [
    path(
        'process/',
        ProcessView.as_view(),
        name='process'
    ),
    re_path(
        r'^response/(?P<status>success|error|cancel)/$',
        ProcessResponseView.as_view(),
        name='process-response'
    ),
    path(
        'transaction-report/',
        TransactionReportView.as_view(),
        name='transaction-report'
    ),
]

if settings.DEBUG:
    urlpatterns += [
        path(
            'test/',
            TestView.as_view(),
            name='test'
        ),
    ]



from django.shortcuts import redirect
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import FormView, View
from django.http.response import HttpResponse

from wspay.conf import settings, resolve
from wspay.forms import (
    UnprocessedPaymentForm, WSPayErrorResponseForm, WSPaySuccessResponseForm,
    WSPayCancelResponseForm, WSPayTransactionReportForm,
)
from wspay.models import WSPayRequestStatus
from wspay.services import (
    process_response_data, process_transaction_report, render_wspay_form,
    verify_response, verify_transaction_report
)
from wspay.signals import process_response_pre_redirect


class ProcessView(FormView):
    """Receive payment data and prepare it for WSPay."""

    form_class = UnprocessedPaymentForm
    template_name = 'wspay/error.html'

    def form_valid(self, form):
        return render_wspay_form(form, self.request)


class PaymentStatus:
    SUCCESS = 'success'
    ERROR = 'error'
    CANCEL = 'cancel'


class ProcessResponseView(View):
    """Handle success, error and cancel."""

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        form_class, request_status, redirect_url = self._unpack_response_status(kwargs['status'])

        data = request.POST if request.method == 'POST' else request.GET
        wspay_request = process_response_data(
            verify_response(form_class, data),
            request_status
        )

        process_response_pre_redirect.send_robust(
            self.__class__,
            wspay_request=wspay_request, http_request=request, redirect_url=redirect_url,
        )

        return redirect(redirect_url)

    def _unpack_response_status(self, status):
        assert status in [PaymentStatus.SUCCESS, PaymentStatus.ERROR, PaymentStatus.CANCEL]
        if status == PaymentStatus.SUCCESS:
            form_class = WSPaySuccessResponseForm
            request_status = WSPayRequestStatus.COMPLETED
            redirect_url = resolve(settings.WS_PAY_SUCCESS_URL)
        elif status == PaymentStatus.CANCEL:
            form_class = WSPayCancelResponseForm
            request_status = WSPayRequestStatus.CANCELLED
            redirect_url = resolve(settings.WS_PAY_CANCEL_URL)
        else:
            form_class = WSPayErrorResponseForm
            request_status = WSPayRequestStatus.FAILED
            redirect_url = resolve(settings.WS_PAY_ERROR_URL)

        return form_class, request_status, redirect_url


class TransactionReportView(View):
    """Transaction response - CallbackURL."""

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        print('dispatch', request.method)
        data = request.POST if request.method == 'POST' else request.GET
        print(data)
        process_transaction_report(
            verify_transaction_report(WSPayTransactionReportForm, data)
        )
        return HttpResponse('OK')


class TestView(FormView):
    """Simple View to test the ProcessView."""

    template_name = 'wspay/test.html'
    form_class = UnprocessedPaymentForm



# Generated by Django 4.0.1 on 2022-01-17 17:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wspay', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='WSPayTransaction',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('payload', models.TextField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.AddField(
            model_name='wspayrequest',
            name='transactions',
            field=models.ManyToManyField(to='wspay.WSPayTransaction'),
        ),
    ]



from appconf import AppConf

from django.conf import settings  # noqa


def resolve(setting):
    """Resolve setting to a result of a callable or itself."""
    if hasattr(setting, '__call__'):
        return setting()
    return setting


class WSPayAppConf(AppConf):

    SUCCESS_URL = '/'
    ERROR_URL = '/'
    CANCEL_URL = '/'
    DEVELOPMENT = None
    VERSION = '2.0'
    SHOP_ID = None
    SECRET_KEY = None

    class Meta:
        prefix = 'ws_pay'
        required = ['SHOP_ID', 'SECRET_KEY']

    def configure_development(self, value):
        value = settings.DEBUG if value is None else value
        return self._configure_potentially_callable_setting(value)

    def configure_shop_id(self, value):
        return self._configure_potentially_callable_setting(value)

    def configure_secret_key(self, value):
        return self._configure_potentially_callable_setting(value)

    def configure_success_url(self, value):
        return self._configure_potentially_callable_setting(value)

    def configure_error_url(self, value):
        return self._configure_potentially_callable_setting(value)

    def configure_cancel_url(self, value):
        return self._configure_potentially_callable_setting(value)

    def _configure_potentially_callable_setting(self, value):
        # If redirect_url setting is a callable return it
        if hasattr(value, '__call__'):
            return value

        # Try resolving a setting to a fully qualified name
        # of a function, return the function object if found
        try:
            modname, part_symbol, attr = value.rpartition('.')
            assert part_symbol == '.', value
            assert modname != '', value
        except Exception:
            return value

        try:
            m = __import__(modname, fromlist=[attr])
            f = getattr(m, attr)
            return f
        except Exception:
            raise



from datetime import date
from decimal import Decimal
from unittest.mock import patch
import pytest
import responses
import requests

from django.urls import reverse
from django.utils import timezone as tz
from django.test.client import RequestFactory, Client
from wspay.conf import settings, resolve

from wspay.forms import UnprocessedPaymentForm, WSPaySignedForm, WSPayTransactionReportForm
from wspay.models import Transaction, WSPayRequest
from wspay import services
from wspay.services import (
    generate_wspay_form_data, generate_signature, get_form_endpoint, get_services_endpoint,
    process_transaction_report, status_check, verify_transaction_report
)

from wspay.tests.utils import STATUS_CHECK_RESPONSE, TRANSACTION_REPORT, MockResponse


def test_incoming_data_form():
    """Test the form that receives cart and user details."""
    form = UnprocessedPaymentForm()
    assert form.is_valid() is False
    form = UnprocessedPaymentForm({'user_id': 1, 'cart_id': 1, 'price': 1})
    assert form.is_valid()


@pytest.mark.django_db
def test_wspay_encode():
    """Test the processing function which prepares the data for WSPay."""
    shop_id = resolve(settings.WS_PAY_SHOP_ID)
    secret_key = resolve(settings.WS_PAY_SECRET_KEY)
    assert shop_id == 'MojShop'
    assert secret_key == '3DfEO2B5Jjm4VC1Q3vEh'

    return_data = {
        'ShopID': shop_id,
        'Version': resolve(settings.WS_PAY_VERSION),
        'TotalAmount': '10,00',
        'ReturnURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'success'})
        ),
        'CancelURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'cancel'})
        ),
        'ReturnErrorURL': (
            'http://testserver' + reverse('wspay:process-response', kwargs={'status': 'error'})
        ),
        'ReturnMethod': 'POST',
    }

    incoming_form = UnprocessedPaymentForm({'cart_id': 1, 'price': 10})
    if (incoming_form.is_valid()):
        form_data = generate_wspay_form_data(
            incoming_form.cleaned_data.copy(), RequestFactory().get('/')
        )

    req = WSPayRequest.objects.get()
    return_data['ShoppingCartID'] = str(req.request_uuid)
    return_data['Signature'] = generate_signature([
        shop_id,
        secret_key,
        str(req.request_uuid),
        secret_key,
        '1000',
        secret_key,
    ])

    assert return_data == form_data


@pytest.mark.django_db
def test_wspay_form():
    """Test the form that is used to make a WSPay POST request."""
    form = WSPaySignedForm()
    assert form.is_valid() is False

    incoming_form = UnprocessedPaymentForm({'user_id': 1, 'cart_id': 1, 'price': 1})
    if (incoming_form.is_valid()):
        form_data = generate_wspay_form_data(
            incoming_form.cleaned_data.copy(),
            RequestFactory().get('/')
        )

    form = WSPaySignedForm(form_data)
    assert form.is_valid()
    form = form.cleaned_data

    responses.add(responses.POST, 'https://formtest.wspay.biz/authorization.aspx', status=200)
    response = requests.post('https://formtest.wspay.biz/authorization.aspx', form)
    assert response.status_code == 200


@pytest.mark.django_db
def test_transaction_update(settings):
    """Test wspay transaction update callback."""
    request = WSPayRequest.objects.create(cart_id=1)
    TRANSACTION_REPORT['ShoppingCartID'] = str(request.request_uuid)
    r = Client().post(
        reverse('wspay:transaction-report'),
        TRANSACTION_REPORT,
    )
    assert r.status_code == 200
    assert r.content == b'OK'

    request.refresh_from_db()
    assert request.transaction_id is not None

    tx = request.transaction
    assert tx.transaction_datetime == tz.datetime(2020, 4, 23, 8, 24, 57, tzinfo=tz.utc)
    assert tx.approval_code == '961792'
    assert tx.ws_pay_order_id == '96a5f58f-764c-4640-90ea-591280893bff'
    assert tx.stan == '38967'
    assert tx.amount == Decimal('78')
    assert tx.history.count() == 1
    assert tx.authorized is True
    assert tx.completed is False
    assert tx.refunded is False
    assert tx.voided is False
    assert tx.can_complete is True
    assert tx.can_void is True
    assert tx.can_refund is False
    assert tx.expiration_date == date(2020, 6, 30)


@pytest.mark.django_db
def test_status_check():
    """Test WSPay status check method."""
    request = WSPayRequest.objects.create(cart_id=1)

    STATUS_CHECK_RESPONSE.update({'ShoppingCartID': str(request.request_uuid)})

    response = MockResponse(status_code=200, json_data=STATUS_CHECK_RESPONSE)
    with patch.object(requests, 'post', return_value=response):
        status_check(request.request_uuid)

    request.refresh_from_db()
    assert request.transaction_id is not None

    tx = request.transaction
    assert tx.transaction_datetime == tz.datetime(2022, 1, 19, 20, 16, 35, tzinfo=tz.utc)
    assert tx.approval_code == '307006'
    assert tx.ws_pay_order_id == '4ae43ae9-af8f-4ddd-8a21-7864f9fe81fb'
    assert tx.stan == '128729'
    assert tx.amount == Decimal('80.55')
    assert tx.history.count() == 1
    assert tx.authorized is True
    assert tx.completed is False
    assert tx.refunded is False
    assert tx.voided is False
    assert tx.can_complete is True
    assert tx.can_void is True
    assert tx.can_refund is False
    assert tx.expiration_date == date(2022, 12, 31)


def test_conf_resolver():
    """Test conf resolver when settings are callables or dotted path to a callable."""
    assert resolve(settings.WS_PAY_SHOP_ID) == 'MojShop'
    assert resolve(settings.WS_PAY_SECRET_KEY) == '3DfEO2B5Jjm4VC1Q3vEh'


def test_get_form_endpoint(settings):
    """Test get_form_endpoint function."""
    assert settings.WS_PAY_DEVELOPMENT is True
    assert get_form_endpoint() == 'https://formtest.wspay.biz/authorization.aspx'

    settings.WS_PAY_DEVELOPMENT = False
    assert get_form_endpoint() == 'https://form.wspay.biz/authorization.aspx'


def test_get_services_endpoint(settings):
    """Test get_services_endpoint setting."""
    assert settings.WS_PAY_DEVELOPMENT is True
    assert get_services_endpoint() == 'https://test.wspay.biz/api/services'

    settings.WS_PAY_DEVELOPMENT = False
    assert get_services_endpoint() == 'https://secure.wspay.biz/api/services'


@pytest.mark.django_db
@pytest.mark.parametrize(['action', 'allow_attr'], [
    ('complete', 'can_complete'),
    ('refund', 'can_refund'),
    ('void', 'can_void')
])
def test_transaction_actions(action, allow_attr):
    """Test `complete`, `refund` and `void` function."""
    request = WSPayRequest.objects.create(
        cart_id=1,
    )
    TRANSACTION_REPORT['ShoppingCartID'] = str(request.request_uuid)
    transaction = process_transaction_report(
        verify_transaction_report(WSPayTransactionReportForm, TRANSACTION_REPORT)
    )
    assert Transaction.objects.count() == 1

    action_func = getattr(services, action)

    if action != 'void':
        with pytest.raises(AssertionError):
            action_func(transaction, Decimal(100))

    setattr(transaction, allow_attr, False)
    transaction.save()
    with pytest.raises(AssertionError):
        action_func(transaction)

    setattr(transaction, allow_attr, True)
    transaction.save()
    json_data = {
        'WsPayOrderId': "c1c7ad40-6293-417e-9dff-8390822efcb6",
        "ShopID": "MYSHOP",
        "ApprovalCode": "559974",
        "STAN": "38400",
        "ErrorMessage": "",
        "ActionSuccess": "1",
        "Signature": ''.join([
            '9f3c7e03bfe3937a1dd82fec8de8719c12ce01679fcf0f96070436b094f7192dab0ff1c8bef',
            '64d65f167049c9df4ca46a12efa1de2a52cba83b20c9cda88ff2f',
        ])
    }
    response = MockResponse(status_code=200, json_data=json_data)
    with patch.object(requests, 'post', return_value=response):
        action_func(transaction)



# Generated by Django 4.0.1 on 2022-01-20 15:12

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wspay', '0003_rename_wspaytransaction_transactionhistory_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='wspayrequest',
            name='transactions',
        ),
    ]



# Generated by Django 4.0.1 on 2022-01-20 15:12

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('wspay', '0002_wspaytransaction_wspayrequest_transactions'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='WSPayTransaction',
            new_name='TransactionHistory',
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('request_uuid', models.UUIDField(default=uuid.uuid4, unique=True)),
                ('stan', models.CharField(max_length=10)),
                ('amount', models.DecimalField(decimal_places=2, max_digits=9)),
                ('approval_code', models.CharField(max_length=10)),
                ('ws_pay_order_id', models.CharField(max_length=50)),
                ('transaction_datetime', models.DateTimeField()),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('authorized', models.BooleanField(default=False)),
                ('completed', models.BooleanField(default=False)),
                ('voided', models.BooleanField(default=False)),
                ('refunded', models.BooleanField(default=False)),
                ('can_complete', models.BooleanField(default=False)),
                ('can_void', models.BooleanField(default=False)),
                ('can_refund', models.BooleanField(default=False)),
                ('expiration_date', models.DateField()),
                ('history', models.ManyToManyField(to='wspay.TransactionHistory')),
            ],
        ),
        migrations.AddField(
            model_name='wspayrequest',
            name='transaction',
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to='wspay.transaction'
            ),
        ),
    ]



from django import forms
from django.core.exceptions import ValidationError


class UnprocessedPaymentForm(forms.Form):
    cart_id = forms.CharField()
    price = forms.DecimalField()
    first_name = forms.CharField(required=False)
    last_name = forms.CharField(required=False)
    address = forms.CharField(required=False)
    city = forms.CharField(required=False)
    zip_code = forms.CharField(required=False)
    country = forms.CharField(required=False)
    email = forms.EmailField(required=False)
    phone = forms.CharField(required=False)


class WSPaySignedForm(forms.Form):
    ShopID = forms.CharField(widget=forms.HiddenInput())
    ShoppingCartID = forms.CharField(widget=forms.HiddenInput())
    Version = forms.CharField(widget=forms.HiddenInput())
    TotalAmount = forms.CharField(widget=forms.HiddenInput())
    Signature = forms.CharField(widget=forms.HiddenInput())
    ReturnURL = forms.CharField(widget=forms.HiddenInput())
    CancelURL = forms.CharField(widget=forms.HiddenInput())
    ReturnErrorURL = forms.CharField(widget=forms.HiddenInput())
    ReturnMethod = forms.CharField(widget=forms.HiddenInput())

    # CUSTOMER DATA
    CustomerFirstName = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerLastName = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerAddress = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerCity = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerZIP = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerCountry = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerEmail = forms.CharField(required=False, widget=forms.HiddenInput())
    CustomerPhone = forms.CharField(required=False, widget=forms.HiddenInput())


class WSPayBaseResponseForm(forms.Form):
    ShoppingCartID = forms.CharField()
    Success = forms.IntegerField()
    ApprovalCode = forms.CharField(required=False)
    Signature = forms.CharField()


class WSPaySuccessErrorResponseForm(WSPayBaseResponseForm):
    CustomerFirstName = forms.CharField(required=False)
    CustomerSurname = forms.CharField(required=False)
    CustomerAddress = forms.CharField(required=False)
    CustomerCity = forms.CharField(required=False)
    CustomerZIP = forms.CharField(required=False)
    CustomerCountry = forms.CharField(required=False)
    CustomerPhone = forms.CharField(required=False)
    CustomerEmail = forms.CharField(required=False)
    Lang = forms.CharField(required=False)
    DateTime = forms.CharField(required=False)
    Amount = forms.CharField()
    ECI = forms.CharField(required=False)
    PaymentType = forms.CharField(required=False)
    PaymentPlan = forms.CharField(required=False)
    ShopPostedPaymentPlan = forms.CharField(required=False)
    ShopPostedLang = forms.CharField(required=False)
    ShopPostedCreditCardName = forms.CharField(required=False)
    ShopPostedPaymentMethod = forms.CharField(required=False)


class WSPaySuccessResponseForm(WSPaySuccessErrorResponseForm):
    STAN = forms.CharField(required=False)
    Partner = forms.CharField(required=False)
    WsPayOrderId = forms.CharField()
    CreditCardNumber = forms.CharField(required=False)
    ApprovalCode = forms.CharField()
    ErrorMessage = forms.CharField(required=False)

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data['Success'] != 1 or cleaned_data['ApprovalCode'] == '':
            raise ValidationError(
                'Expecting success to be 1 and approval code to not be blank.'
            )


class WSPayErrorResponseForm(WSPaySuccessErrorResponseForm):
    ErrorMessage = forms.CharField()
    ErrorCodes = forms.CharField()


class WSPayCancelResponseForm(WSPayBaseResponseForm):
    ResponseCode = forms.CharField()


class WSPayTransactionReportForm(forms.Form):
    WsPayOrderId = forms.CharField()

    # Declared as deprecated by WSPay docs
    UniqueTransactionNumber = forms.CharField(required=False)

    Signature = forms.CharField()
    STAN = forms.CharField(required=False)
    ApprovalCode = forms.CharField()
    ShopID = forms.CharField()
    ShoppingCartID = forms.CharField()
    Amount = forms.CharField()
    CurrencyCode = forms.CharField()
    ActionSuccess = forms.IntegerField()

    # Declared as deprecated by WSPay docs
    Success = forms.IntegerField(required=False)

    Authorized = forms.IntegerField()
    Completed = forms.IntegerField()
    Voided = forms.IntegerField()
    Refunded = forms.IntegerField()
    PaymentPlan = forms.CharField(required=False)
    Partner = forms.CharField(required=False)
    OnSite = forms.IntegerField(required=False)
    CreditCardName = forms.CharField(required=False)
    CreditCardNumber = forms.CharField(required=False)
    ECI = forms.CharField(required=False)
    CustomerFirstName = forms.CharField(required=False)
    CustomerSurname = forms.CharField(required=False)
    CustomerAddress = forms.CharField(required=False)
    CustomerCity = forms.CharField(required=False)
    CustomerZIP = forms.CharField(required=False)
    CustomerCountry = forms.CharField(required=False)
    CustomerPhone = forms.CharField(required=False)
    CustomerEmail = forms.CharField(required=False)
    TransactionDateTime = forms.CharField()
    IsLessThen30DaysFromTransaction = forms.BooleanField(required=False)
    CanBeCompleted = forms.BooleanField(required=False)
    CanBeVoided = forms.BooleanField(required=False)
    CanBeRefunded = forms.BooleanField(required=False)
    Token = forms.CharField(required=False)
    TokenNumber = forms.CharField(required=False)
    ExpirationDate = forms.CharField(required=False)



from enum import Enum
import uuid

from django.db import models
from django.utils.translation import gettext_lazy as _


class WSPayRequestStatus(Enum):
    PENDING = 'pending'
    COMPLETED = 'completed'
    CANCELLED = 'cancelled'
    FAILED = 'failed'

    @classmethod
    def choices(cls):
        return [(x.name, x.value) for x in cls]


class TransactionHistory(models.Model):
    payload = models.TextField()
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)


class Transaction(models.Model):
    request_uuid = models.UUIDField(default=uuid.uuid4, unique=True)
    stan = models.CharField(max_length=10)
    amount = models.DecimalField(max_digits=9, decimal_places=2)
    approval_code = models.CharField(max_length=10)
    ws_pay_order_id = models.CharField(max_length=50)
    transaction_datetime = models.DateTimeField()
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)
    history = models.ManyToManyField(to=TransactionHistory)
    authorized = models.BooleanField(default=False)
    completed = models.BooleanField(default=False)
    voided = models.BooleanField(default=False)
    refunded = models.BooleanField(default=False)
    can_complete = models.BooleanField(default=False)
    can_void = models.BooleanField(default=False)
    can_refund = models.BooleanField(default=False)
    expiration_date = models.DateField()


class WSPayRequest(models.Model):
    cart_id = models.PositiveIntegerField()
    status = models.CharField(
        max_length=15, choices=WSPayRequestStatus.choices(),
        default=WSPayRequestStatus.PENDING.name
    )
    request_uuid = models.UUIDField(default=uuid.uuid4)
    response = models.TextField(null=True, blank=False)
    additional_data = models.TextField(
        null=False,
        blank=True,
        help_text=_('Use this to store any data you want to preserve when making a request')
    )
    transaction = models.OneToOneField(
        to=Transaction, null=True, blank=True, on_delete=models.PROTECT
    )
    created = models.DateTimeField(auto_now_add=True)
    updated = models.DateTimeField(auto_now=True)



import calendar
from decimal import setcontext, Decimal, BasicContext
from datetime import datetime, date
from enum import Enum
import json
import hashlib
from typing import Tuple
import requests
import pytz

from django.core.exceptions import ValidationError
from django.shortcuts import render
from django.urls import reverse

from wspay.conf import settings, resolve
from wspay.forms import WSPaySignedForm, WSPayTransactionReportForm
from wspay.models import Transaction, WSPayRequest, TransactionHistory, WSPayRequestStatus
from wspay.signals import pay_request_created, pay_request_updated

EXP = Decimal('.01')
setcontext(BasicContext)


def render_wspay_form(form, request, additional_data=''):
    """
    Render the page that will submit signed data to wspay.

    Convert input data into WSPay format
    Generate wspay signature

    Return an HttpResponse that will submit the form data to wspay.
    """
    if not form.is_valid():
        raise ValidationError(form.errors)
    wspay_form = WSPaySignedForm(
        generate_wspay_form_data(form.cleaned_data.copy(), request, additional_data)
    )
    return render(
        request,
        'wspay/wspay_submit.html',
        {'form': wspay_form, 'submit_url': get_form_endpoint()}
    )


def generate_wspay_form_data(input_data, request, additional_data=''):
    """Process incoming data and prepare for POST to WSPay."""
    wspay_request = WSPayRequest.objects.create(
        cart_id=input_data['cart_id'],
        additional_data=additional_data,
    )
    # Send a signal
    pay_request_created.send_robust(WSPayRequest, instance=wspay_request)

    input_data['cart_id'] = str(wspay_request.request_uuid)

    price = input_data['price']
    assert price > 0, 'Price must be greater than 0'
    total_for_sign, total = build_price(price)

    shop_id = resolve(settings.WS_PAY_SHOP_ID)
    secret_key = resolve(settings.WS_PAY_SECRET_KEY)
    signature = generate_signature([
        shop_id,
        secret_key,
        input_data['cart_id'],
        secret_key,
        total_for_sign,
        secret_key,
    ])

    return_data = {
        'ShopID': shop_id,
        'ShoppingCartID': input_data['cart_id'],
        'Version': resolve(settings.WS_PAY_VERSION),
        'TotalAmount': total,
        'Signature': signature,
        'ReturnURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'success'})
        ),
        'CancelURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'cancel'})
        ),
        'ReturnErrorURL': request.build_absolute_uri(
            reverse('wspay:process-response', kwargs={'status': 'error'})
        ),
        'ReturnMethod': 'POST',
    }
    if input_data.get('first_name'):
        return_data['CustomerFirstName'] = input_data['first_name']
    if input_data.get('last_name'):
        return_data['CustomerLastName'] = input_data['last_name']
    if input_data.get('address'):
        return_data['CustomerAddress'] = input_data['address']
    if input_data.get('city'):
        return_data['CustomerCity'] = input_data['city']
    if input_data.get('zip_code'):
        return_data['CustomerZIP'] = input_data['zip_code']
    if input_data.get('country'):
        return_data['CustomerCountry'] = input_data['country']
    if input_data.get('email'):
        return_data['CustomerEmail'] = input_data['email']
    if input_data.get('phone'):
        return_data['CustomerPhone'] = input_data['phone']

    return return_data


def verify_response(form_class, data):
    """Verify validity and authenticity of wspay response."""
    form = form_class(data=data)
    if form.is_valid():
        signature = form.cleaned_data['Signature']
        shop_id = resolve(settings.WS_PAY_SHOP_ID)
        secret_key = resolve(settings.WS_PAY_SECRET_KEY)
        param_list = [
            shop_id,
            secret_key,
            data['ShoppingCartID'],
            secret_key,
            data['Success'],
            secret_key,
            data['ApprovalCode'],
            secret_key,
        ]
        expected_signature = generate_signature(param_list)
        if signature != expected_signature:
            raise ValidationError('Bad signature')

        return form.cleaned_data

    raise ValidationError('Form is not valid')


def verify_transaction_report(form_class, data):
    """Verify validity and authenticity of wspay transaction report."""
    form = form_class(data=data)
    if form.is_valid():
        signature = form.cleaned_data['Signature']
        shop_id = resolve(settings.WS_PAY_SHOP_ID)
        secret_key = resolve(settings.WS_PAY_SECRET_KEY)
        param_list = [
            shop_id,
            secret_key,
            form.cleaned_data['ActionSuccess'],
            form.cleaned_data['ApprovalCode'],
            secret_key,
            shop_id,
            form.cleaned_data['ApprovalCode'],
            form.cleaned_data['WsPayOrderId'],
        ]
        expected_signature = generate_signature(param_list)
        if signature != expected_signature:
            raise ValidationError('Bad signature')

        return form.cleaned_data

    raise ValidationError('Form is not valid')


def process_response_data(response_data, request_status):
    """Update corresponding WSPayRequest object with response data."""
    wspay_request = WSPayRequest.objects.get(
        request_uuid=response_data['ShoppingCartID'],
    )
    wspay_request.status = request_status.name
    wspay_request.response = json.dumps(response_data)
    wspay_request.save()

    if wspay_request.status == WSPayRequestStatus.COMPLETED.name:
        status_check(wspay_request.request_uuid)

    # Send a signal
    pay_request_updated.send_robust(
        WSPayRequest,
        instance=wspay_request,
        status=request_status
    )

    return wspay_request


def process_transaction_report(response_data):
    """Create a transaction and append to relevant wspay request."""
    request_uuid = response_data['ShoppingCartID']
    wspay_request = WSPayRequest.objects.get(
        request_uuid=request_uuid,
    )

    transaction_datetime = pytz.timezone(
        'Europe/Zagreb'
    ).localize(
        datetime.strptime(response_data['TransactionDateTime'], '%Y%m%d%H%M%S')
    )

    expires = datetime.strptime(response_data['ExpirationDate'], '%y%m').date()
    (_, day) = calendar.monthrange(expires.year, expires.month)
    expires = date(expires.year, expires.month, day)

    transaction, created = Transaction.objects.update_or_create(
        request_uuid=request_uuid,
        defaults={
            'stan': response_data['STAN'],
            'amount': Decimal(response_data['Amount']),
            'approval_code': response_data['ApprovalCode'],
            'ws_pay_order_id': response_data['WsPayOrderId'],
            'transaction_datetime': transaction_datetime,
            'authorized': bool(response_data['Authorized']),
            'completed': bool(response_data['Completed']),
            'voided': bool(response_data['Voided']),
            'refunded': bool(response_data['Refunded']),
            'can_complete': bool(response_data['CanBeCompleted']),
            'can_void': bool(response_data['CanBeVoided']),
            'can_refund': bool(response_data['CanBeRefunded']),
            'expiration_date': expires
        }
    )
    if created:
        wspay_request.transaction = transaction
        wspay_request.save()

    # TODO: Update status
    transaction_history = TransactionHistory.objects.create(
        payload=json.dumps(response_data)
    )
    transaction.history.add(transaction_history)

    # TODO: Send a signal

    return transaction


def status_check(request_uuid):
    """Check status of a transaction."""
    version = '2.0'
    shop_id = resolve(settings.WS_PAY_SHOP_ID)
    secret_key = resolve(settings.WS_PAY_SECRET_KEY)
    shopping_cart_id = str(request_uuid)
    signature = generate_signature([
        shop_id,
        secret_key,
        shopping_cart_id,
        secret_key,
        shop_id,
        shopping_cart_id
    ])

    data = {
        'Version': version,
        'ShopId': shop_id,
        'ShoppingCartId': shopping_cart_id,
        'Signature': signature
    }

    r = requests.post(
        f'{get_services_endpoint()}/statusCheck',
        data=data
    )
    return process_transaction_report(
        verify_transaction_report(WSPayTransactionReportForm, r.json())
    )


class TransactionAction(Enum):
    Complete = 'completion'
    Refund = 'refund'
    Void = 'void'


def complete(transaction: Transaction, amount: Decimal = None) -> Tuple[bool, str]:
    """Complete preauthorized transaction."""
    assert transaction.can_complete
    assert amount is None or amount <= transaction.amount
    return _transaction_action(
        transaction,
        int((amount or transaction.amount) * 100),
        TransactionAction.Complete
    )


def refund(transaction: Transaction, amount: Decimal = None) -> Tuple[bool, str]:
    """Refund refundable transaction."""
    assert transaction.can_refund
    assert amount is None or amount <= transaction.amount
    return _transaction_action(
        transaction,
        int((amount or transaction.amount) * 100),
        TransactionAction.Refund
    )


def void(transaction: Transaction) -> Tuple[bool, str]:
    """Void voidable transaction."""
    assert transaction.can_void
    return _transaction_action(
        transaction,
        int(transaction.amount * 100),
        TransactionAction.Void
    )


def _transaction_action(
    transaction: Transaction, amount: int, action: TransactionAction
) -> Tuple[bool, str]:
    version = '2.0'
    shop_id = resolve(settings.WS_PAY_SHOP_ID)
    secret_key = resolve(settings.WS_PAY_SECRET_KEY)

    signature = generate_signature([
        shop_id,
        transaction.ws_pay_order_id,
        secret_key,
        transaction.stan,
        secret_key,
        transaction.approval_code,
        secret_key,
        str(amount),
        secret_key,
        transaction.ws_pay_order_id
    ])

    data = {
        'Version': version,
        'WsPayOrderId': transaction.ws_pay_order_id,
        'ShopID': shop_id,
        'ApprovalCode': transaction.approval_code,
        'STAN': transaction.stan,
        'Amount': amount,
        'Signature': signature
    }
    r = requests.post(
        f'{get_services_endpoint()}/{action.value}',
        data=data
    )
    response_data = r.json()

    expected_signature = generate_signature([
        shop_id,
        secret_key,
        response_data['STAN'],
        response_data['ActionSuccess'],
        secret_key,
        response_data['ApprovalCode'],
        response_data['WsPayOrderId'],
    ])

    if response_data['Signature'] != expected_signature:
        raise ValidationError('Bad signature')

    # Per WSPay docs update approval_code in case it was changed
    new_approval_code = response_data['ApprovalCode']
    if transaction.approval_code != new_approval_code:
        transaction.approval_code = new_approval_code
        transaction.save()

    success = bool(int(response_data['ActionSuccess']))
    error_message = response_data.get('ErrorMessage', '')
    return success, error_message


def generate_signature(param_list):
    """Compute the signature."""
    result = []
    for x in param_list:
        result.append(str(x))
    return compute_hash(''.join(result))


def compute_hash(signature):
    """Compute the hash out of the given values."""
    return hashlib.sha512(signature.encode()).hexdigest()


def build_price(price):
    """
    Round to two decimals and return the tuple containing two variations of price.

    First element of the tuple is an int repr of price as as str 123.45 => '12345'
    Second element is a str that is a properly formatted price 00123.451 => '123,45'
    """
    rounded = price.quantize(EXP)
    _, digits, exp = rounded.as_tuple()

    result = []
    digits = list(map(str, digits))
    build, next = result.append, digits.pop

    for i in range(2):
        build(next() if digits else '0')
    build(',')
    if not digits:
        build('0')

    while digits:
        build(next())

    return str(int(rounded * 100)), ''.join(reversed(result))


def get_form_endpoint():
    """Return production or dev endpoint based on DEVELOPMENT setting."""
    development = resolve(settings.WS_PAY_DEVELOPMENT)
    if development:
        return 'https://formtest.wspay.biz/authorization.aspx'
    return 'https://form.wspay.biz/authorization.aspx'


def get_services_endpoint():
    """Return production or dev services endpoint based on DEVELOPMENT setting."""
    development = resolve(settings.WS_PAY_DEVELOPMENT)
    if development:
        return 'https://test.wspay.biz/api/services'
    return 'https://secure.wspay.biz/api/services'



DEBUG = True
USE_TZ = True
TIME_ZONE = "UTC"
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "HOST": "127.0.0.1",
        "NAME": "wspay",
        "USER": ""
    }
}
MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]
ROOT_URLCONF = "wspay.tests.urls"
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "wspay",
]
SITE_ID = 1
TEMPLATES = [{
    "BACKEND": "django.template.backends.django.DjangoTemplates",
    "APP_DIRS": True,
    "OPTIONS": {
        "debug": True,
        "context_processors": [
            "django.contrib.auth.context_processors.auth",
            "django.template.context_processors.debug",
            "django.template.context_processors.i18n",
            "django.template.context_processors.media",
            "django.template.context_processors.static",
            "django.template.context_processors.tz",
            "django.template.context_processors.request",
        ],
    },
}]
SECRET_KEY = "ws-pay-secret-key"


def _shop_id():
    return 'MojShop'


WS_PAY_SHOP_ID = _shop_id
WS_PAY_SECRET_KEY = 'wspay.tests.utils.secret_key'
WS_PAY_DEVELOPMENT = True



TRANSACTION_REPORT = {
    'WsPayOrderId': '96a5f58f-764c-4640-90ea-591280893bff',
    'UniqueTransactionNumber': 46370,
    'Signature': ''.join([
        'b77be564829eda566d516d544b08e6ff16968ce13c5566781ae12621c685819b14ae1fea124a87',
        'fa1e72b94127a91cd319982b065b4a4d490ff875e6dc115f93',
    ]),
    'STAN': '38967',
    'ApprovalCode': '961792',
    'ShopID': 'MYSHOP',
    'ShoppingCartID': '1c731772-e407-45f8-b576-7e20b0e8642d',
    'Amount': 78,
    'CurrencyCode': 191,
    'ActionSuccess': '1',
    'Success': '1',
    'Authorized': '1',
    'Completed': '0',
    'Voided': '0',
    'Refunded': '0',
    'PaymentPlan': '0000',
    'Partner': 'Pbz',
    'OnSite': '1',
    'CreditCardName': 'AMEX',
    'CreditCardNumber': '377500*****1007',
    'ECI': '',
    'CustomerFirstName': 'John',
    'CustomerLastName': 'Doe',
    'CustomerAddress': 'Street address 10',
    'CustomerCity': 'City',
    'CustomerCountry': 'HR',
    'CustomerPhone': '0911111111111111',
    'CustomerZIP': '51000',
    'CustomerEmail': 'john@doe.com',
    'TransactionDateTime': '20200423102457',
    'IsLessThen30DaysFromTransaction': True,
    'CanBeCompleted': True,
    'CanBeVoided': True,
    'CanBeRefunded': False,
    'Token': '09bf3d95-7a57-43f3-873a-216467a39925',
    'TokenNumber': '1007',
    'ExpirationDate': '2006'
}

STATUS_CHECK_RESPONSE = {
    'WsPayOrderId': '4ae43ae9-af8f-4ddd-8a21-7864f9fe81fb',
    'UniqueTransactionNumber': 380037,
    'Signature': ''.join([
        '7a266ceb2abad8001fa33a8dac5a270a47a716521bab40cc88e3158843fbf70d176dc832462',
        'bf904436ed3d3060b30623a76af3d040918ab005108249dd962d0',
    ]),
    'STAN': '128729',
    'ApprovalCode': '307006',
    'ShopID': 'LJEKPLUS',
    'ShoppingCartID': '5f9c3535-0c25-4f6b-a86a-819ca1b27eec',
    'Amount': 80.55,
    'CurrencyCode': 191,
    'ActionSuccess': '1',
    'Success': '1',
    'Authorized': '1',
    'Completed': '0',
    'Voided': '0',
    'Refunded': '0',
    'PaymentPlan': '0000',
    'Partner': 'Pbz',
    'OnSite': '1',
    'CreditCardName': 'VISA',
    'CreditCardNumber': '478561******0189',
    'ECI': '',
    'CustomerFirstName': 'Vedran',
    'CustomerLastName': 'Vojvoda',
    'CustomerAddress': 'Testna ulica 1',
    'CustomerCity': 'Zagreb',
    'CustomerCountry': 'HR',
    'CustomerPhone': '55512345',
    'CustomerZIP': '10000',
    'CustomerEmail': 'vedran@pinkdroids.com',
    'TransactionDateTime': '20220119211635',
    'IsLessThen30DaysFromTransaction': True,
    'CanBeCompleted': True,
    'CanBeVoided': True,
    'CanBeRefunded': False,
    'ExpirationDate': '2212',
}


def secret_key():
    """Resolve secret key setting."""
    return '3DfEO2B5Jjm4VC1Q3vEh'


class MockResponse:
    def __init__(self, status_code, json_data):
        self.json_data = json_data
        self.status_code = status_code

    def json(self):
        return self.json_data
